/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.fleet.vehicle;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.fleet.model.*;
import nl.cloudfarming.client.model.*;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectNotFoundException;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import org.openide.util.lookup.ServiceProviders;

/**
 *
 * @author Wytse Visser
 */
@ServiceProviders({
    @ServiceProvider(service = VehicleManager.class),
    @ServiceProvider(service = EntityManager.class)
})
public class VehicleManagerImpl extends AbstractEntityManager<Vehicle> implements VehicleManager, MachineManager<Vehicle>, MachineProvider {

    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.fleet.vehicle");
    private static final String FLEET_DIR = "Fleet";
    private static final String VEHICLES_DIR = "Vehicles";

    @Override
    public Vehicle load(FileObject fileObject) {
        Vehicle vehicle = null;
        try (ObjectInputStream ois = new ObjectInputStream(fileObject.getInputStream())) {
            vehicle = (Vehicle) ois.readObject();
        } catch (ClassNotFoundException | IOException | NullPointerException ex) {
            LOGGER.log(Level.WARNING, "Could not create vehicle from fileObject {0}", ex.getMessage());
        }
        return getCachedInstanceFor(vehicle);
    }

    @Override
    public void save(Vehicle vehicle) {
        try {
            FarmManager farmManager = Lookup.getDefault().lookup(FarmManager.class);
            Project project = farmManager.findFarmProject(vehicle.getURI());
            FileObject vehiclesFolder = getVehiclesSubFolder(project, true);
            FileObject vehicleObject = vehiclesFolder.getFileObject(vehicle.getURI().getItemIdNumber(), Vehicle.EXT);
            if (vehicleObject == null) {
                vehicleObject = vehiclesFolder.createData(vehicle.getURI().getItemIdNumber(), Vehicle.EXT);
            }
            save(vehicle, vehicleObject);

        } catch (IOException ex) {
            LOGGER.log(Level.SEVERE, "Error while trying to create file object for vehicle {0}", ex.getMessage());
        }

    }

    private synchronized void save(Vehicle vehicle, FileObject fileObject) {
        try (ObjectOutputStream oos = new ObjectOutputStream(fileObject.getOutputStream())) {
            oos.writeObject(vehicle);
            addToCache(vehicle);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "Could not save vehicle to file {0} {1}", new Object[]{vehicle.getName(), ex.getMessage()});
        }
    }

    private static FileObject getFleetFolder(Project project, boolean create) {
        FileObject farmFolder = project.getProjectDirectory();
        FileObject fleetFolder = farmFolder.getFileObject(FLEET_DIR);

        if (fleetFolder == null && create) {
            try {
                fleetFolder = farmFolder.createFolder(FLEET_DIR);
            } catch (IOException ex) {
                LOGGER.log(Level.SEVERE, "Could not create the folder {0} in folder {1}", new Object[]{FLEET_DIR, farmFolder.getPath()});
            }
        }
        return fleetFolder;
    }

    private static FileObject getVehiclesSubFolder(Project project, boolean create) {
        FileObject fleetFolder = getFleetFolder(project, create);
        FileObject vehiclesFolder = fleetFolder.getFileObject(VEHICLES_DIR);

        // Create folder if it doesn't exist
        if (vehiclesFolder == null && create) {
            try {
                vehiclesFolder = fleetFolder.createFolder(VEHICLES_DIR);
            } catch (IOException ex) {
                LOGGER.log(Level.SEVERE, "Could not create the folder {0} in folder {1}", new Object[]{VEHICLES_DIR, fleetFolder.getPath()});
            }
        }
        return vehiclesFolder;
    }

    @Override
    public Set<ItemIdType> getSupportedTypes() {
        return EnumSet.of(ItemIdType.VCE);
    }

    @Override
    protected Vehicle load(AgroURI uri, Project project) {
        switch (uri.getItemIdType()) {
            case VCE:
                FileObject fileObject = getVehiclesSubFolder(project, true).getFileObject(uri.getItemIdNumber(), Vehicle.EXT);
                return load(fileObject);
            default:
                LOGGER.log(Level.WARNING, "cannot load unsupported entity type {0}", new Object[]{uri.getItemIdType()});
                return null;
        }
    }

    @Override
    public void remove(Vehicle entity) {
        //TODO implement remove
        throw new UnsupportedOperationException("Not supported yet.");
    }

    /**
     * Method that finds all vehicles of a farm. It loops through all the
     * FileObjects contained by the vehicles folder, gets a DataObject for each
     * FileObject, fetches the Vehicle object from the DataObject and uses the
     * find method of the AbstractEntityManager to either fetch the vehicle from
     * the cache or load it using the proper channels. Using the cache prevents
     * vehicles from being loaded multiple times.
     *
     * @param project The project you wish to fetch all vehicles from
     * @return List containing all the vehicles of the project passed as
     * parameter
     */
    @Override
    public List<Machine> findAll(Project project) {
        List<Machine> vehicles = new ArrayList<>();
        FileObject vehiclesDir = getVehiclesSubFolder(project, true);

        //for each serialized vehicle file
        for (FileObject vehicleFile : vehiclesDir.getChildren()) {
            try {
                //get the DataObject for the file
                DataObject dataObject = DataObject.find(vehicleFile);
                if (dataObject != null) {
                    //get the Vehicle contained by the DataObject
                    Vehicle vehicle = dataObject.getLookup().lookup(Vehicle.class);
                    if (vehicle != null) {
                        //use the AbstractEntityManager to get the Vehicle from the cache
                        //if possible
                        vehicles.add(find(vehicle.getURI()));
                    }
                }
            } catch (DataObjectNotFoundException ex) {
                LOGGER.log(Level.WARNING, "Could not find DataObject for file {0} {1}", new Object[]{vehicleFile, ex.getMessage()});
            }
        }

        return vehicles;
    }
}
